
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/vnode.h>
#include <sys/mount.h>
#include <sys/malloc.h>

#include "ods2.h"

int ods2_decode_mapping_pointer __P((unsigned char *mp,
                                     unsigned int *lbn,
                                     unsigned int *count,
                                     unsigned int *type,
                                     unsigned int *mp_size));
int convert_to_physical __P((struct ods2_node *ip, daddr_t vbn, daddr_t *lbn));

int
ods2_decode_mapping_pointer( unsigned char *mp,
                             unsigned int *lbn,
                             unsigned int *count,
                             unsigned int *type,
                             unsigned int *mp_size )
{
    struct fm2_1_t *fm1;
    struct fm2_2_t *fm2;
    struct fm2_3_t *fm3;

    *type = MAPPING_TYPE(mp);

    switch (*type)
    {
        case FM2_C_PLACEMENT :
            *mp_size = FM2_S_PLACEMENT;
            break;

        case FM2_C_FORMAT1 :
            fm1 = (struct fm2_1_t *) mp;
            *mp_size = FM2_S_FORMAT1;
            *lbn = LBN_1(*fm1);
            *count = COUNT_1(*fm1) + 1;
            break;

        case FM2_C_FORMAT2 :
            fm2 = (struct fm2_2_t *) mp;
            *mp_size = FM2_S_FORMAT2;
            *lbn = LBN_2(*fm2);
            *count = COUNT_2(*fm2) + 1;
            break;

        case FM2_C_FORMAT3 :
            fm3 = (struct fm2_3_t *) mp;
            *mp_size = FM2_S_FORMAT3;
            *lbn = LBN_3(*fm3);
            *count = COUNT_3(*fm3) + 1;
            break;

        default :
            printf("Unknown mapping pointer type");
            return EINVAL;
    }

    return 0;
}

int
convert_to_physical( struct ods2_node *ip, daddr_t vbn, daddr_t *plbn )
{
    unsigned char *mp;

    unsigned int bytes_in_use;
    unsigned int lbn;
    unsigned int count;
    unsigned int type;
    unsigned int mp_size;
    unsigned int n;
    unsigned int blocks_into_file;
    unsigned int in_extension;
    int error;
    ino_t extension_file_number;
    struct file_header_t *fh2;

    in_extension = 0;
    blocks_into_file = 0;
    fh2 = ip->file_header;

    do
    {
        bytes_in_use = fh2->fh2_b_map_inuse * 2;
        mp = FM2_AREA(fh2);

        for (n=0; n < bytes_in_use;)
        {
            error = ods2_decode_mapping_pointer(mp,
                                                &lbn,
                                                &count,
                                                &type,
                                                &mp_size);

            if (error)
            {
                if (in_extension)
                    FREE( fh2, M_ODS2FSFH );

                return error;
            }

            n += mp_size;

            if (type != FM2_C_PLACEMENT)
            {
                if (vbn <= blocks_into_file + count)
                {
                    /*
                     * VBN = 1 => first LBN.
                     */

                    if (in_extension)
                        FREE( fh2, M_ODS2FSFH );

                    *plbn = lbn + vbn - blocks_into_file - 1;

                    return 0;
                }

                blocks_into_file += count;
            }

            mp += mp_size;
        }

        extension_file_number = FILE_NUMBER(fh2->fh2_w_ext_fid);
        if ( extension_file_number != 0 )
        {
            /*
             * The file may have more than one extension.  We need to free the
             * memory allocated by ods2_read_file_header each time we get to
             * here, except for the first time.  The first time, in_extension
             * will be false.
             */

            if ( in_extension )
                FREE( fh2, M_ODS2FSFH );

            in_extension = 1;

            error = ods2_read_file_header( ip->mount,
                                           extension_file_number,
                                           &fh2 );
            if ( error )
                return error;
        }
    } while ( extension_file_number != 0 );

    /* Should not happen if read, etc. did their job of checking eof */

    return EINVAL;
}

int
ods2_bmap(ap)
	struct vop_bmap_args /* {
		struct vnode *a_vp;
		daddr_t  a_bn;
		struct vnode **a_vpp;
		daddr_t *a_bnp;
		int *a_runp;
		int *a_runb;
	} */ *ap;
{
        int error;
	struct ods2_node *ip = VTOODS2(ap->a_vp);
	daddr_t vbn = ap->a_bn;

        if (vbn <= 0)
            return(EINVAL);

if (0)
printf("ods2_bmap[starting]\n");

	/*
	 * Check for underlying vnode requests and ensure that logical
	 * to physical mapping is requested.
	 */

	if (ap->a_vpp != NULL)
		*ap->a_vpp = ip->i_devvp;
	if (ap->a_bnp == NULL)
		return (0);

	/*
	 * Compute the requested block number
	 */

        error = convert_to_physical(ip, vbn, ap->a_bnp);
        if ( error )
            return error;

if (0)
printf("ods2_bmap[%d -> %d]\n", vbn, *ap->a_bnp);

	/*
	 * Determine maximum number of readahead blocks following the
	 * requested block.  Not yet implemented.  Must read ahead based
         * on the current mapping pointer.
	 */
/*
	if (ap->a_runp) {
		int nblk;

		nblk = (ip->i_size >> bshift) - (vbn + 1);
		if (nblk <= 0)
			*ap->a_runp = 0;
		else if (nblk >= (MAXBSIZE >> bshift))
			*ap->a_runp = (MAXBSIZE >> bshift) - 1;
		else
			*ap->a_runp = nblk;
	}
*/
        if (ap->a_runp)
            *ap->a_runp = 0;

	if (ap->a_runb)
		*ap->a_runb = 0;

	return 0;
}
