
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/vnode.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/fcntl.h>
#include <sys/conf.h>
#include <sys/buf.h>

#include "ods2.h"

int
ods2_bread(struct vnode *vp,
           struct ods2_geometry *geometry,
           daddr_t lbn,
           struct ucred *cred,
           unsigned char *block)
{
    int error;
    unsigned block_offset;
    daddr_t sector;
    struct buf *bp;

    /*
     * bread knows the sector size of the media.  It wants reads to be
     * sector aligned AND to be an integral number of sectors.  Therefore,
     * we round down the lbn and read a full sector.  We then have to
     * extract our block from the sector.
     */

    /*
     * Round the lbn down to sector boundary
     */

    sector = lbn & ~geometry->block_mask;

    /*
     * Calculate where the block falls within the sector
     */

    block_offset = lbn & geometry->block_mask;
    block_offset *= ODS2_BLOCK_SIZE;

    if (0)
        printf("lbn=%d, sector=%d, shift=%d, mask=%x, block_offset=%d\n",
               lbn, sector, geometry->block_shift,
               geometry->block_mask, block_offset);

    error = bread(vp, sector, geometry->sector_size, cred, &bp);
    if (error)
    {
        brelse(bp);
        return error;
    }

    /*
     * Move the correct block
     */

    bcopy(bp->b_data + block_offset, block, ODS2_BLOCK_SIZE);

    brelse(bp);

    return 0;
}
